/*
  KeePass Password Safe - The Open-Source Password Manager
  Copyright (C) 2003-2019 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;

using KeePass.App;
using KeePass.Resources;
using KeePass.UI;
using KeePass.Util;

using KeePassLib;
using KeePassLib.Keys;
using KeePassLib.Native;
using KeePassLib.Utility;
using KeePass;
using KeePassLib.Security;
using System.Security.Cryptography;
using System.Reflection;

namespace WinKee.Forms
{
    public partial class WinKeeForm : Form
    {
        //tb 06012012 ProtectedBinary m_pKey = new ProtectedBinary(true);
        ProtectedBinary m_pKey = new ProtectedBinary(true, new byte[128]);

        private bool m_bCanExit = false;
        private bool m_bHasExited = false;

        private bool m_bInitializing = false;

        public byte[] Entropy = { 1, 2, 3, 5, 7, 11, 13, 17 };
        public string EncryptedPassword { get; set; }
        public string EncryptedKeyPath { get; set; }
        public bool AutoRun { get; set; }

        public ProtectedBinary Key
        {
            get
            {
                Debug.Assert(m_pKey != null);
                return m_pKey;
            }
        }

        public bool HasClosedWithExit
        {
            get { return m_bHasExited; }
        }

        public WinKeeForm()
        {
            InitializeComponent();
            SecureTextBoxEx.InitEx(ref m_tbPassword);
            EncryptedPassword = "";
            EncryptedKeyPath = "";
            AutoRun = false;
            Program.Translation.ApplyTo(this);
        }

        private void OnFormLoad(object sender, EventArgs e)
        {

            GlobalWindowManager.AddWindow(this);

            var versionAttribute = Assembly.GetExecutingAssembly().GetCustomAttributes(typeof(AssemblyFileVersionAttribute), true).FirstOrDefault() as AssemblyFileVersionAttribute;
            lblREV.Text = versionAttribute.Version;

            m_bInitializing = true;

            m_ttRect.SetToolTip(m_cbHidePassword, KPRes.TogglePasswordAsterisks);
            m_ttRect.SetToolTip(m_btnOpenKeyFile, KPRes.KeyFileSelect);

            m_tbPassword.Text = string.Empty;
            m_tbPassword.TextChanged += this.ProcessTextChangedPassword;
            if (EncryptedPassword != null && EncryptedPassword.Length > 0)
            {
                byte[] encrypted = Convert.FromBase64String(EncryptedPassword);
                byte[] decrypted = ProtectedData.Unprotect(encrypted, Entropy,
                    DataProtectionScope.CurrentUser);
                m_tbPassword.TextEx = new ProtectedString(true, decrypted);
                Array.Clear(decrypted, 0, decrypted.Length);
            }

            m_tbKeyPath.Text = KPRes.NoKeyFileSpecifiedMeta;
            m_cbKeyFile.Checked = false;
            if (EncryptedKeyPath != null && EncryptedKeyPath.Length > 0)
            {
                byte[] encrypted = Convert.FromBase64String(EncryptedKeyPath);
                byte[] decrypted = ProtectedData.Unprotect(encrypted, Entropy,
                    DataProtectionScope.CurrentUser);
                m_tbKeyPath.Text = Encoding.UTF8.GetString(decrypted);
                Array.Clear(decrypted, 0, decrypted.Length);
                m_cbKeyFile.Checked = true;
            }

            m_cbHidePassword.Checked = true;
            OnCheckedHidePassword(sender, e);

            Debug.Assert(m_tbKeyPath.Text.Length != 0);

            m_btnExit.Enabled = m_bCanExit;
            m_btnExit.Visible = m_bCanExit;

            CustomizeForScreenReader();
            EnableUserControls();

            m_bInitializing = false;

            this.BringToFront();
            this.Activate();
            m_tbPassword.Focus();

            if (AutoRun) this.m_btnOK.PerformClick();
        }

        private void CustomizeForScreenReader()
        {
            if (!Program.Config.UI.OptimizeForScreenReader) return;

            m_cbHidePassword.Text = KPRes.HideUsingAsterisks;
            m_btnOpenKeyFile.Text = KPRes.SelectFile;
        }

        private void CleanUpEx()
        {
            m_tbPassword.TextChanged -= this.ProcessTextChangedPassword;
        }

        private bool ComputeKey()
        {
            MemoryStream ms = new MemoryStream();

            if (m_cbPassword.Checked) // Use a password
            {
                try
                {
                    byte[] decrypted = m_tbPassword.TextEx.ReadUtf8();
                    byte[] encrypted = ProtectedData.Protect(decrypted, Entropy,
                        DataProtectionScope.CurrentUser);
                    EncryptedPassword = Convert.ToBase64String(encrypted);
                    Array.Clear(encrypted, 0, encrypted.Length);
                    KcpPassword kcpPw = new KcpPassword(decrypted);
                    Array.Clear(decrypted, 0, decrypted.Length);

                    byte[] pbKeyData = kcpPw.KeyData.ReadData();
                    ms.Write(pbKeyData, 0, pbKeyData.Length);
                    Array.Clear(pbKeyData, 0, pbKeyData.Length);
                    //tb 06012012 kcpPw.Clear();
                    kcpPw = null;
                }
                catch (Exception)
                {
                    MessageService.ShowWarning(KPRes.Password, KPRes.ObjectNotFound);
                    return false;
                }
            }

            string strKeyFile = m_tbKeyPath.Text;
            Debug.Assert(strKeyFile != null); if (strKeyFile == null) strKeyFile = string.Empty;
            if (m_cbKeyFile.Checked && (!strKeyFile.Equals(KPRes.NoKeyFileSpecifiedMeta)))
            {
                if (ValidateKeyFileLocation() == false) return false;
                try
                {
                    byte[] decrypted = UnicodeEncoding.UTF8.GetBytes(m_tbKeyPath.Text);
                    byte[] encrypted = ProtectedData.Protect(decrypted, Entropy,
                        DataProtectionScope.CurrentUser);
                    Array.Clear(decrypted, 0, decrypted.Length);

                    EncryptedKeyPath = Convert.ToBase64String(encrypted);
                    Array.Clear(encrypted, 0, encrypted.Length);

                    KcpKeyFile kcpFile = new KcpKeyFile(m_tbKeyPath.Text);
                    byte[] pbKeyData = kcpFile.KeyData.ReadData();
                    ms.Write(pbKeyData, 0, pbKeyData.Length);
                    Array.Clear(pbKeyData, 0, pbKeyData.Length);

                    //tb 06012012 kcpFile.Clear();
                    kcpFile = null;

                }
                catch (Exception e)
                {
                    MessageService.ShowWarning(strKeyFile, KPRes.KeyFileError, e.Message);
                    return false;
                }
            }

            byte[] finalKey = ms.ToArray();
            if (finalKey.Length == 0) return false;
            m_pKey = new ProtectedBinary(true, finalKey);
            Array.Clear(finalKey, 0, finalKey.Length);

            return true;
        }

        private bool ValidateKeyFileLocation()
        {
            string strKeyFile = m_tbKeyPath.Text;
            Debug.Assert(strKeyFile != null);
            if (strKeyFile == null) strKeyFile = string.Empty;
            if (strKeyFile.Equals(KPRes.NoKeyFileSpecifiedMeta)) return true;

            bool bSuccess = true;

            if (File.Exists(strKeyFile) == false)
            {
                MessageService.ShowWarning(strKeyFile, KPRes.FileNotFoundError);
                bSuccess = false;
            }

            return bSuccess;
        }

        private void EnableUserControls()
        {
            string strKeyFile = m_tbKeyPath.Text;
            Debug.Assert(strKeyFile != null);
            if (strKeyFile == null) strKeyFile = string.Empty;
            if (m_cbKeyFile.Checked && strKeyFile.Equals(KPRes.NoKeyFileSpecifiedMeta))
                m_btnOK.Enabled = false;
            else m_btnOK.Enabled = true;
        }

        private void OnCheckedPassword(object sender, EventArgs e)
        {
            if (m_cbPassword.Checked) m_tbPassword.Focus();
        }

        private void OnCheckedKeyFile(object sender, EventArgs e)
        {
            if (m_bInitializing) return;
            if (!m_cbKeyFile.Checked) m_tbKeyPath.Text = "";
            EnableUserControls();
        }

        private void ProcessTextChangedPassword(object sender, EventArgs e)
        {
            m_cbPassword.Checked = (m_tbPassword.Text.Length != 0);
        }

        private void OnCheckedHidePassword(object sender, EventArgs e)
        {
            m_tbPassword.EnableProtection(m_cbHidePassword.Checked);
            //something to think about - do we want to hide the file path (???)
            //tb: m_tbKeyPath.UseSystemPasswordChar = m_tbPassword.UseSystemPasswordChar;
        }

        private void OnBtnOK(object sender, EventArgs e)
        {
            if (!ComputeKey()) this.DialogResult = DialogResult.None;
            else CleanUpEx();
        }

        private void OnBtnCancel(object sender, EventArgs e)
        {
            m_pKey = null;
            CleanUpEx();
        }

        private void OnClickKeyFileBrowse(object sender, EventArgs e)
        {
            string strFilter = UIUtil.CreateFileTypeFilter("key", KPRes.KeyFiles, true);
            //OpenFileDialog ofd = UIUtil.CreateOpenFileDialog(KPRes.KeyFileSelect, strFilter, 2, null, false, true);
            OpenFileDialogEx ofd = UIUtil.CreateOpenFileDialog(KPRes.KeyFileSelect, strFilter, 2, null, false, AppDefs.FileDialogContext.KeyFile);


            if (ofd.ShowDialog() == DialogResult.OK)
            {
                m_cbKeyFile.Checked = true;

                m_tbKeyPath.Text = ofd.FileName;
            }

            EnableUserControls();
        }

        private void OnFormClosed(object sender, FormClosedEventArgs e)
        {
            GlobalWindowManager.RemoveWindow(this);
        }

        private void OnBtnExit(object sender, EventArgs e)
        {
            if (m_bCanExit == false)
            {
                Debug.Assert(false);
                this.DialogResult = DialogResult.None;
                return;
            }
            m_bHasExited = true;
        }
    }
}
